#!/usr/bin/env python
'''
    Given input with individual variants over a region of the genome, 
    generate an interval within those variants that passes the 
    four-gamete filtering criteria, then return either that interval
    or an output file with variants in that interval.

    The four-gamete test is a method for determining whether or not 
    there has been recombination between a pair of variants. To do this, 
    all individuals have haplotypes defined as the variants at the two
    sites. Given two snps with ref/alt alleles A/G and C/T, if 
    individuals in this sample have haplotypes AC, AT, and GT, it is
    possible that there has been no recombination between these alleles.
    If an additional individual has the GC haplotype, this means that 
    a recombination event must have taken place between the sites. 

    This test must be done on phased data.

    Common usage for this function is to input a VCF file that contains
    variants for individuals at a single locus, with output returned being
    a VCF that contains a subsample of these variants. A full VCF can be 
    used with --vcfreg, where the second argument is a BED file with one
    or more regions, output will be either a VCF for four-gamete passing 
    regions or a new BED file with the truncated regions.

    ###############
    Input Arguments
    ###############
    **--vcfs** *<input_vcf_1>*...*<input_vcf_n>*
        Input name of one or more VCF files, where each VCF represents
        a locus. 
    **--vcfreg** *<input_vcf>* *<BED file>*
        Input name of VCF file containing genome data and name of BED file
        with regions to be analyzed.

    ###############
    Output Aguments
    ###############
    **--out** *<output_filename>*
        Name for output file.
    **--out-prefix** *<ouput_prefix>*
        If multiple files are output, this option is required to set a 
        prefix for the output files.

    ##################
    Interval Arguments
    ##################
    **--numinf** *<minimum informative site count>*
        Region returned must have at least n informative sites, defaults to 1
    **--hk**
        If set, returns intervals with at least one recombination event 
        instead of regions with no recombination.
    **--reti**
        This script will generate a list of valid regions with no
        recombination. Selecting this option will return a single interval
        as specified by other arguments
    **--retl**
        Returns all valid intervals, either as a list of intervals or 
        multiple output files


    ################################
    Single Returned Region Arguments
    ################################
    Select one of:
    **--rani**
        Returns random interval (default)
    **--ranb**
        Returns random interval, with probability of interval 
        proportional to interval length
    **--left**
        Return first interval with enough informative sites
    **--right**
        Return last interval with enough informative sites
    **--maxlen**
        Return interval with most informative sites

    ###############
    Other Arguments
    ###############
    **--remove-multiallele**
        Removes multi-alleleic sites from analysis
    **--include-missing**
        Include sites with missing data in analysis
    **--ovlps**
        Extend region to include non-informative variants between
        an edge variant and a variant that breaks the four-gamete
        criteria
    **--ovlpi**
        Include informative variants from overlapping
        regions
'''

import sys
import argparse
import logging
import random
import pysam

from pgpipe.logging_module import initLogger
from pgpipe.genome_region import Region, RegionList
from pgpipe.vcf_reader_func import getRecordList, vcfRegionName, getRecordsInRegion, VcfReader


class BaseData():

    def __init__(self, args, format, filename, region=None):
        self.poslist = []
        self.seqs = []
        self.format = format
        self.header = None
        self.records = []
        self.extend_noninf = args.extend_noninf
        self.extend_inf = args.extend_inf
        if format == 'VCF':
            self.onlysnps = True
            vcff = VcfReader(filename,index=args.tabix_index)
            self.allrecords = vcff.getRecordList(region=region, chrom=args.chrom)
            self.checkVcfRegion()
            self.header = vcff.reader.header
            vcff.close()
        elif format == "FASTA":
            self.onlysnps = False
            f = open(filename,"r")
            seq = ""
            while True:
                s = f.readline().strip()
                if len(s) == 0:
                    if len(seq) > 0:
                        self.seqs.append(seq)
                    break
                else:
                    if s[0] == '>':
                        if len(seq) > 0:
                            self.seqs.append(seq)
                        seq = ""
                    else:
                        s = s.strip()
                        if len(s) > 0:
                            seq += s
            f.close()
        elif format == "SIT":
            self.onlysnps = False
            f = open(filename,"r")
            f.readline()
            while True:
                s = f.readline()
                if s[0] != '#':
                    break
            numseq = int(s.split()[0])
            seqlen = int(s.split()[1])
            s = f.readline().split()

            numnoncode = int(s[1])
            if numnoncode != seqlen:
                for i in range(numnoncode):
                    f.readline()
            s = f.readline()
            if s.find("All") < 0:
                numpops = int(s.strip())
                for i in range(numpops):
                    f.readline()
            ## now should be at beginning of data
            for i in range(numseq):
                self.seqs.append(f.readline().strip()[10:])
            f.close()
        self.numbases = listcheckallsamelength(self.seqs)
        logging.info('Sequence length: %d' % self.numbases)
        self.seqcount = len(self.seqs)
        logging.info('Sequence count: %d' % self.seqcount)
        self.buildlistofsites(args.includesnpmissingdata, args.only2basesnps)

    def buildlistofsites(self, includesnpmissingdata, only2basesnps):
        MAXINFORMSNPS = 1000
        # Find polymorphic sites
        self.polysites = []
        self.informpolysites = []
        informpolyseqs = ['']*self.seqcount ## becomes a list of strings of informative sites,  same length as seqs,  not used as of 8/14/2017
        baseset = set(['a','c','g','t','A','C','G','T'])
        linesets = []
        for j in range(self.numbases):
            ok = True
            basecount = [0,0,0,0]
            for k in range(self.seqcount):
                if self.seqs[k][j].isalpha() == False:
                    raise Exception ("non-alphabetic character %s"
                        " at position %d in sequence %d"%(self.seqs[k][j],k,j))
                if self.seqs[k][j] not in baseset: ## skip sites with missing data
                    if not includesnpmissingdata:
                        ok = False
                if self.seqs[k][j].upper() == 'A' :
                    basecount[0] = basecount[0] + 1
                if self.seqs[k][j].upper() == 'C' :
                    basecount[1] = basecount[1] + 1
                if self.seqs[k][j].upper() == 'G' :
                    basecount[2] = basecount[2] + 1
                if self.seqs[k][j].upper() == 'T' :
                    basecount[3] = basecount[3] + 1
            if ok == False:
                pass
                # print ("missing data at position " + str(j))
            else:
                polybase = 0
                twobase = 0
                for bi in range(4):
                    if basecount[bi] > 0 :
                        polybase += 1
                        if basecount[bi] > 1 :
                            twobase +=  + 1
                if  only2basesnps == True and polybase > 2 :
                    pass
                    # print ("more than two bases at position " + str(j))
                elif polybase > 1:
                    self.polysites.append(j+1) ## use regular (start at 1) count
                    if (twobase > 1):
                        self.informpolysites.append(j+1)
                        lineset =[set([]),set([]),set([]),set([])]
                        for k in range(self.seqcount):
                            if self.seqs[k][j].upper() == 'A' :
                                lineset[0].add(k)
                            if self.seqs[k][j].upper() == 'C' :
                                lineset[1].add(k)
                            if self.seqs[k][j].upper() == 'G' :
                                lineset[2].add(k)
                            if self.seqs[k][j].upper() == 'T' :
                                lineset[3].add(k)
                        linesets.append(lineset)
                        if len(self.informpolysites) >= MAXINFORMSNPS:
                            logging.warning('Maximum SNP count of %d has been hit'%(MAXINFORMSNPS))
                            break
        logging.info('Informative site count: %d' % len(self.informpolysites))
        c = 0
        sl = len(linesets)
        if sl > 1 :
            self.cmat = []
            self.cintervals = []
            for i in range(sl):
                self.cmat.append([])
                for j in range(sl):
                    if i < j :
                        if self.records[i].chrom != self.records[j].chrom or compat(linesets[i],linesets[j]) == False :
                            self.cmat[i].append(0)
                            self.cintervals.append([self.informpolysites[i],self.informpolysites[j]])
                        else:
                            self.cmat[i].append(1)
                    else:
                        if i==j:
                            self.cmat[i].append(1)
                        else:
                            self.cmat[i].append(-1)


    def compatibleintervals(self):
        """
         return a list of intervals
            CONTIG4GPASS, set of most inclusive intervals, such that all the sites
                in each are fully compatible with each other:
                each interval is a stretch of bases that contain zero incompatible
                sites by the 4 gamete criterion
                on each side,  the left and right,  the next polymorphic site that
                is not in the interval must be incompatible
                with one of the polymorphic sites that is in the interval.

                intervals are based only on informative sites
                it is possible for intervals could be extended to include flanking
                non-informative polymorphic sites
                probably would have little effect in most contexts
        """
        tempinformpolysites = list(self.informpolysites)
        cints = []
        nps = len(self.informpolysites)
        cintinformlistfull = []
        p = 0
        while p < nps:
            j = p
            donej = False
            cintinform = 1
            while True:
                j += 1
                if j >= nps or self.cmat[p][j] != 1:
                    j -= 1
                    donej = True
                    break
                else:
                    for i in range(p+1,j+1):
                        if self.cmat[i][j] != 1:
                            donej = True
                            j -= 1
                            break
                if donej:
                    break
                else:
                    cintinform += 1
            # if j > p:
            cints.append([tempinformpolysites[p],tempinformpolysites[j]])
            cintinformlistfull.append(cintinform)
            p += 1
        ## now remove all intervals that are overlapped by a larger one
        cintinformremovelist = []
        c = len(cints)
        removelist = []
        for ai in range(c-1) :
            for bi in range(ai+1,c) :
                assert (cints[ai][0] <= cints[ai][1])
                assert (cints[bi][0] <= cints[bi][1])
                i = cints[ai][0]
                j = cints[ai][1]
                m = cints[bi][0]
                n = cints[bi][1]
                if (m <= i)  and (j <= n):
                    if cints[bi] not in removelist:
                        removelist.append(cints[ai])
                        cintinformremovelist.append(ai)
                if (i <= m)  and (n <= j):
                    if cints[ai] not in removelist:
                        removelist.append(cints[bi])
                        cintinformremovelist.append(bi)
        for interval in removelist:
            cints.remove(interval)
        cintinformlist = []
        for i in range(c):
            if i not in cintinformremovelist:
                cintinformlist.append(cintinformlistfull[i])
        # now pad intervals with sequences to the flanking ones
        cintspadded = []
        if False:
            cints = replace_positions(cints,self.poslist)
            return cints,cintinformlist
        if self.onlysnps:
        ## using a vcf file,  intervals are 1-based  but positions in list_of_positions are 0-based
            if not self.extend_inf:
                for ci in range(len(cints)):
                    tempcint = [-1,-1]
                    if ci == 0:
                        tempcint[0] = cints[ci][0]
                    else:
                        tempcint[0] = self.informpolysites[self.informpolysites.index(cints[ci-1][1])+1]
                    tempcint[1] = cints[ci][1]
                    cints[ci] = tempcint
                    cintinformlist[ci] = self.informpolysites.index(cints[ci][1]) - self.informpolysites.index(cints[ci][0])+1
            for ci in range(len(cints)):
                tempcint = [-1,-1]
                if not self.extend_noninf:
                    tempcint[0] = self.poslist[cints[ci][0]-1]
                    if ci == len(cints)-1:
                        tempcint[1] = self.poslist[-1]
                    else:
                        tempcint[1] = self.poslist[self.informpolysites[self.informpolysites.index(cints[ci][1])+1]-1]-1
                else:
                    if ci == 0:
                        assert cints[ci][0] > 0
                        tempcint[0] = self.poslist[cints[ci][0] - 1]  ## -1 because list_of_positions is 0-based
                    else:
                        tempcint[0] = self.poslist[self.informpolysites[self.informpolysites.index(cints[ci][0]) -1] -1] + 1  ## position of previous informative snp + 1
                    if ci == len(cints)-1:
                        tempcint[1] = self.poslist[cints[ci][1] -1]
                    else:
                        tempcint[1] = self.poslist[self.informpolysites[self.informpolysites.index(cints[ci][1]) + 1] -1] -1   ## position of next informative snp -1
                cintspadded.append(tempcint)
        else:
            tempcint = [-1,-1]
            if not self.extend_inf:
                for ci in range(len(cints)):
                    if ci == 0:
                        tempcint[0] = cints[ci][0]
                    else:
                        tempcint[0] = self.informpolysites[self.informpolysites.index(cints[ci-1][1])+1]
                    tempcint[1] = cints[ci][1]
                    cints[ci] = tempcint
                    cintinformlist[ci] = self.informpolysites.index(cints[ci][1]) - self.informpolysites.index(cints[ci][0])+1
            else:
                for ci in range(len(cints)):
                    if ci == 0:
                        tempcint[0] = 1
                    else:
                        tempcint[0] = self.informpolysites[self.informpolysites.index(cints[ci][0]) -1] + 1   ## position of previous informative snp + 1
                    if ci == len(cints)-1:
                        tempcint[1] = self.numbases
                    else:
                        tempcint[1] = self.informpolysites[self.informpolysites.index(cints[ci][1]) + 1] -1    ## position of next informative snp -1
                    cintspadded.append(tempcint)
        return cintspadded,cintinformlist
        #return cints,cintinformlist

    #def hudsonkaplan85intervals(cintervals,numbases,list_of_positions) :
    def hk85intervals(self):
        """
            implements the HK algorithm from appendix 2 of Hudson & Kaplan 1985

            Rm is the number of recombination events that can be
            parsimoniously inferred from a sample of sequences
            D is a compatibility matrix d[i,j] == 1 if all 4 gametes are present,
                 0 otherwise
            consider for convenience only i<j i.e. above the diagonal

            H&K algorithm prunes D to leave the minimum set of intervals
            necessary to explain the data

            pruning steps:
                for two intervals (i,j) and (m,n)  if m <= i< j <= n  remove (m,n)

                for remaining intervals do these steps to remove
                non disjoint intervals
                    identify an interval that is not disjoint from all
                    the others (if none can be found we are done)
                    call this (i1,j1)

                    now scan all intervals (m,n) and delete
                    all those for which i1 < m < j2

                    now search other intervals to find (i2,j2) such that i2 >= j1
                    and this interval is not disjoint from all the
                    remaining intervals
                    now using (i2,j2) delete all intervals
                    whose first component is <j2 and >i2
        """
        c = len(self.cintervals)
        removelist = []
        for ai in range(c-1) :
            for bi in range(ai+1,c) :
                assert (self.cintervals[ai][0] < self.cintervals[ai][1])
                assert (self.cintervals[bi][0] < self.cintervals[bi][1])
                i = self.cintervals[ai][0]
                j = self.cintervals[ai][1]
                m = self.cintervals[bi][0]
                n = self.cintervals[bi][1]
                if (m <= i)  and (j <= n):
                    if self.cintervals[bi] not in removelist:
                        removelist.append(self.cintervals[bi])
                if (i <= m)  and (n <= j):
                    if self.cintervals[ai] not in removelist:
                        removelist.append(self.cintervals[ai])
        for interval in removelist:
            self.cintervals.remove(interval)
        c = len(self.cintervals)
        ai = 0
        while ai < c-1:
            removelist = []
            for bi in range(ai+1,c):
                i1 = self.cintervals[ai][0]
                j1 = self.cintervals[ai][1]
                b = self.cintervals[bi][0]
                e = self.cintervals[bi][1]
                if  (i1 < b  < j1) or (b < i1 < e):
                    removelist.append(self.cintervals[bi])
            for interval in removelist:
                self.cintervals.remove(interval)

            i1 = self.cintervals[ai][0]
            j1 = self.cintervals[ai][1]
            xi = ai + 1
            c = len(self.cintervals)
            while xi < c:
                if self.cintervals[xi][0] >= j1:
                    removelist = []
                    for bi in range(xi+1,c):
                        i2 = self.cintervals[xi][0]
                        j2 = self.cintervals[xi][1]
                        b = self.cintervals[bi][0]
                        e = self.cintervals[bi][1]
                        if  (i2 < b  < j2):
                            removelist.append(self.cintervals[bi])
                for interval in removelist:
                    self.cintervals.remove(interval)
                    c = len(self.cintervals)
                xi += 1
            c = len(self.cintervals)
            ai += 1
        # reset snp positions if needed
        if len(self.poslist) > 0:
            self.cintervals = replace_positions(self.cintervals,self.poslist)
        return self.cintervals

    def checkVcfRegion(self):
        """
        Checks that variants in region are SNPs and creates a list with valid sites. Also sets ploidy and position list of valid sites.

        """
        valid_bases = ['A','C','G','T']
        skiplines = []
        self.ploidy = -1
        self.poslist = []
        base_list = []
        self.records = []
        for record in self.allrecords:
            lineok=True
            for allele in record.alleles:
                if allele not in valid_bases:
                    lineok=False
                    break
            if not lineok:
                continue
            somedata = False
            #data_present = 0
            var_list = []
            #for indiv in record.samples:
            for i in range(len(record.samples)):
                indiv = record.samples[i]
                if self.ploidy == -1:
                    self.ploidy = len(indiv.alleles)
                if not somedata and indiv.alleles.count(None) == 0:
                    somedata = True
                for samp_allele in indiv.alleles:
                    if samp_allele is None:
                        var_list.append('N')
                    else:
                        var_list.append(samp_allele)
            if somedata:
                base_list.append(var_list)
                self.poslist.append(record.pos)
                self.records.append(record)
        self.seqs = [[base_list[row][col] for row in range(0, len(base_list))] for col in range(0, len(base_list[0]))]


def createParser():
    parser = argparse.ArgumentParser(description=("Given a file of aligned"
            "sequences or variable sites, will return intervals based"
            " on 4 gamete tests"))
    #parser.add_argument("filename", help="Input filename")
    filetype_group = parser.add_mutually_exclusive_group(required=True)
    filetype_group.add_argument("--vcfreg", dest = "vcfreg", nargs=2,
            help = "VCF file and region list, in that order")
    filetype_group.add_argument("--vcfs", dest="vcfname", nargs="+",
                                help=("One or more input VCF files"))
    filetype_group.add_argument("--fasta", dest = "fasta",
            nargs="+", help = "fasta file(s) of"
            " multiple aligned sequences, all the same length")
    filetype_group.add_argument("--sit", dest = "sit",
            nargs="+", help = " SITES format file(s)")
    output_group = parser.add_mutually_exclusive_group()
    output_group.add_argument("--out", dest="out",help=("Output filename "
                            "for single input"))
    output_group.add_argument("--out-prefix", dest="out_prefix",help=("Prefix "
                            "for multi-file output"))
    intervaltype_group = parser.add_mutually_exclusive_group(required=
            "--reti" in sys.argv)
    intervaltype_group.add_argument("--hk", dest="intervaltype",
            action='store_const', const="HandK85",help = "get a set of "
            "parsimonious intervals each containing at least 1 recombination"
            " event.  Follows appendix 2 of Hudson and Kaplan 1985")
    intervaltype_group.add_argument("--4gcompat", dest="intervaltype",
            action='store_const', const="CONTIG4GPASS",help = "get a set of "
            " intervals such that for each interval all the included bases are"
            " compatible with each other based on the 4 gamete test")
    returntype_group = parser.add_mutually_exclusive_group(required=True)
    returntype_group.add_argument("--reti", dest="returntype",
            action='store_const', const="returninterval",help = " return single "
            " interval")
    returntype_group.add_argument("--retl", dest="returntype",
            action='store_const', const="returnlist", help = " return a list of"
            " all intervals")
    returntype_group = parser.add_mutually_exclusive_group(required=False)
    returntype_group.add_argument("--rani", dest="intervalpicktype",
            action='store_const', const="randominterval",default =
            "leftinterval", help = "return a randomly selected interval"
            " from the list of all intervals")
    returntype_group.add_argument("--ranb", dest="intervalpicktype",
            action='store_const', const="randombase", default="leftinterval",
            help = "return a single randomly selected interval with probabiliy "
            "of selection proportional to interval length")
    returntype_group.add_argument("--left", dest="intervalpicktype",
            action='store_const', const="leftinterval", default="leftinterval",
            help = "return the leftmost interval")
    returntype_group.add_argument("--right", dest="intervalpicktype",
            action='store_const', const="rightinterval",
            default="leftinterval",help = "return the rightmost interval")
    returntype_group.add_argument("--maxlen", dest="intervalpicktype",action="store_const", const="maxregion",default="leftinterval",help="return interval with most SNPs")
    parser.add_argument("--numinf",dest= "numinformativesites",
            action="store",nargs = 1,type = int, default = None,help =
            "limit the picking of an interval to only those with at least this"
            " many informative snps")
    parser.add_argument("--remove-multiallele",dest="only2basesnps",action="store_true",
            default=False,help = "causes snps with more than 2 bases "
            "to be ignored, default is to include them")
    parser.add_argument("--include-missing", dest="includesnpmissingdata",
            action="store_true", default=False,help="causes snps with missing"
            " data to be included, default is to ignore them")
    parser.add_argument("--ranseed", dest="ranseed", type=int,
            action="store",nargs = 1,default = None, help="positive integer to"
            " serve as the random number seed "
            "for use with --ranb or --rani, not required but"
            " useful when debugging")
    parser.add_argument("--chrom", dest="chrom", help="Select variants "
            "from a single specified chromosome")
    parser.add_argument("--ovlps",dest="extend_noninf",action="store_true",
                        help=("Extend region to overlapping non-informative sites"))
    parser.add_argument("--ovlpi",dest="extend_inf",action="store_true",
                        help="Extend region to overlapping informative sites")
    #add checks for correct input type
    parser.add_argument('--tbi', dest="tabix_index",help=("Filepath for tabix "
                        "index file if using a single bgzipped VCF"))
    return parser

def logArgs(args):
    logging.info('Arguments for vcf_region_write:')
    for k in vars(args):
        logging.info('Argument %s: %s' % (k, vars(args)[k]))

def compat(s1, s2):
    """
    s1 and s2 are lists of sets of line numbers, each representing a particular
    polymorphic site.
    Each list as 4 sets,  one for each base: A,C,G & T.
    Each set is all the line  numbers that share that particular base.
    So then a polymorphic site represented by s1 will have
    two or more nonempty sets in s1.
    Same for s2

    compat() counts the number of pairs of sets, betweeen s1 and s2,
    for all possible pairs, for which the size of the intersection is nonzero
    and less than the size of either set. If this is 4 or more then the
    two sites are not compatible

    this should work for 2,3 or 4 polymorphic bases in either site

    """
    count = 0
    for i1 in range(len(s1)):
        for i2 in range(len(s2)):
            if len(s1[i1]) > 1 and len(s2[i2]) > 1:
                intersect = s1[i1]  & s2[i2]
                lenintersect = len(intersect)
                if (lenintersect > 0 and lenintersect < len(s1[i1])
                    and lenintersect < len(s2[i2])):
                    count = count + 1
    if count >= 4 :
        return False
    else :
        return True

def listcheckallsamelength(lcheck):
    """
        takes a list of lists or strings and checks to see that
        they are all the same length returns the length that is shared by all
    """
    i = 0
    for x in lcheck:
        try:
            lval = len(x)
        except TypeError:
            print('position %d not compatible with len()'%i)
        if i>0:
            assert lval == lastlval
        lastlval = lval
        i+=1
    return lval




def replace_positions(intervals,list_of_positions):
    """
        with vcf files the base number of snp locations is contained
        in list_of_positions and the value in intervals[] are simply
        the index positions of the snps in the vcf files.

        This function rewrites the intervals to reflect the original
        snp base positions
    """
    bintervals = []
    for iv in intervals:
        #logging.info('Index: %d %d' % (iv[0],iv[1]))
        #logging.info('Length: %d' % len(list_of_positions))
        bintervals.append([list_of_positions[iv[0]-1],list_of_positions[iv[1]-1]+1])
    return bintervals


def sampleinterval(picktype,numinf,intervals,infcounts):
    """
        picktype : randominterval, randombase,left,right
        numinf : integer or None
        intervals : list of intervals
        infcounts: list of informative site counts,  or None
    """
    if len(intervals) == 0:
        return None

    #if picktype == "leftinterval":
    #    return intervals[0]
    #if picktype == "rightinterval":
    #    return intervals[-1]
    if numinf != None:
        numinf = numinf[0]
    if numinf != None and infcounts != None:
        okints = [i for i in range(len(infcounts)) if infcounts[i] >= numinf ]
    else:
        okints = [i for i in range(len(intervals))]
    n = len(okints)
    if n == 0: # no intervals with that many informative sites
        return None
    if picktype == "leftinterval":
        return intervals[okints[0]]
    if picktype == "rightinterval":
        return intervals[okints[-1]]
    if picktype == "randominterval":
        return  intervals[ okints[ random.randint(0,n-1)]]
    if picktype == "randombase":
        sumbase = 0
        for i in range(n):
            sumbase += intervals[okints[i]][1] -  intervals[okints[i]][0] + 1
        rbn = random.randint(1,sumbase)
        i = 0
        while True:
            if rbn > intervals[okints[i]][1] -  intervals[okints[i]][0] + 1:
                rbn -= intervals[okints[i]][1] -  intervals[okints[i]][0] + 1
                i += 1
            else:
                break
        return  intervals[ okints[i]]
    if picktype == "maxregion":
        #print (okints)
        #print (infcounts)
        #print (intervals)
        return intervals[infcounts.index(max(infcounts))]
    assert False  # should not get here
    return None

def getIntervalList(args, basedata):
    """
    Returns list of intervals from either hk85 or 4gcompat function.
    Also controls return based on args provided (list or single interval)
    """

    if args.intervaltype == "HandK85":
        #hk code doesn't return counts of the informatics sites in each interval
        intervals = basedata.hk85intervals()
        if args.numinformativesites != None:
            raise Exception("--numinf not implemented when using --hk")
        intervalinformativecounts = None # not implemented  for hk
        numinf = None # set to None, hk use of # inform sites not implemented

    else:
        intervals, intervalinformativecounts = basedata.compatibleintervals()

    if args.returntype == 'returnlist':
        return intervals
    else:
        interval = sampleinterval(args.intervalpicktype,
                                  args.numinformativesites,
                                  intervals,intervalinformativecounts)
        return interval


def outputSubregion(args, interval, basedata, region=None, filename=None):
    #create new region from interval
    #find records in range of new interval
    #open file with prefix and region name
    #write records, close file
    """
    Outputs records in identified sub-interval to desired output file
    """
    if interval is None:
        return
    if region is None:
        subregion = Region(interval[0]-1,interval[1],basedata.records[0].chrom)
    else:
        subregion = Region(interval[0]-1,interval[1],region.chrom)
    subrecords = getRecordsInRegion(subregion, basedata.records)
    if filename is None:
        subfn = vcfRegionName(args.out_prefix,subregion,"vcf.gz")
    else:
        subfn = filename
    #print (interval)
    #print (subfn)
    subf = pysam.VariantFile(subfn, 'w', header=basedata.header)
    for record in subrecords:
        subf.write(record)
    subf.close()





def sample_fourgametetest_intervals(sys_args):
    """Returns interval(s) from aligned sequences or snps.

    Given a set of aligned sequences or a vcf file, this function
    returns intervals that have been identified based on four-gamete
    incombpatibility.   Intervals can either flank a recombination event or
    be sure not to contain an incombatibility.


    Parameters
    ----------

    Input (one required):
    --fasta : str
        Filename(s) of input FASTA file for testing
    --vcfreg : str, str
        Filename of input VCF file and BED file with list of
        regions to run tests on
    --vcfs : str
        Filename(s) of VCF files, each corresponding to a region, that will be subsampled for compatible intervals
    --sit : str
        Filename(s) of input SITES format files for testing

    Output types:
    --out : str
        Name of single output file.
    --out-prefix : str
        Prefix for name of output files
    --out-reg : str
        Name of file that stores passing interval per interval in input BED file (currently not working)

    Tests (one required)
    --hk : bool
        Will find intervals that contain at least one recombination event
    --4gcompat : bool
        Will find intervals with zero recombination events in them

    Return options:
    --reti : bool
        Output will be of a single region, indicated by rani/ranb/left/right
    --retl : bool
        Output will be of all passing regions

    For single region:
    --left : bool (default)
        Single output region will be left-most region with enough informative sites
    --right : bool
        Output region will be right-most region
    --rani : bool
        Region will be randomly selected, with each region having equal chance of selection
    --ranb : bool
        Region will be randomly selected, with each region having selection odds proportional to size
    --maxlen : bool
        Region returned is that of one with most informative SNPs

    General options:
    --numinf : int
        Number of sites in an identified interval required for reporting (default is 1)
    --incN : bool
        Include SNPs with missing data (default is to ignore them)
    --ranseed : int
        Seed for picking random interval from list
    --tbi : str
        Path to tabix index for input VCF file if it's name doesn't match
    --ovlps : bool
        Extend region to overlapping non-informative sites
    --ovlpi : bool
        Extend region to overlapping informative sites

    """
    parser = createParser()
    # parser.print_help()
    if len(sys_args) == 0:
        parser.print_help()
        sys.exit(1)
    args = parser.parse_args(sys_args)
    logArgs(args)
    #argdic = vars(args)
    interval_list = []
    multiple_regions = False
    region_output = False
    if args.ranseed != None:
        random.seed(int(args.ranseed[0]))
    filetype = ''
    out_list = (args.out is None and args.out_prefix is None)

    if args.vcfreg is not None:
        filetype = "VCF"
        vcfname = args.vcfreg[0]
        regname = args.vcfreg[1]
        if regname != '-':
            region_list = RegionList(filename=regname)
            #region_output = True
            multiple_regions = (len(region_list.regions) > 1)
            if multiple_regions and not out_list and args.out is not None:
                raise Exception(("VCF with multiple regions cannot have "
                                "single output file (--out)"))
            for region in region_list.regions:
                basedata = BaseData(args, "VCF",vcfname, region)
                intervals = getIntervalList(args, basedata)
                interval_list.append(intervals)
                if args.out_prefix is not None:
                    if args.returntype == 'returnlist':
                        for iv in intervals:
                            outputSubregion(args, iv, basedata)
                    else:
                        outputSubregion(args, intervals, basedata)
                elif args.out is not None:
                    if args.returntype == 'returnlist':
                        raise Exception('Multiple outputs directed to single file')
                    outputSubregion(args, intervals, basedata, region=region, filename=args.out)

        else:
            basedata = BaseData(args, "VCF",vcfname)
            intervals = getIntervalList(args,basedata)
            interval_list.append(intervals)
            if args.out_prefix is not None:
                if args.returntype == 'returnlist':
                    for iv in intervals:
                        outputSubregion(args, iv, basedata)
                else:
                    outputSubregion(args, intervals, basedata)
            elif out_list:
                return interval_list
            elif args.out is not None:
                if args.returntype == 'returnlist':
                    raise Exception(('If --retl is specified, output must not '
                                    'be --out'))
                outputSubregion(args, intervals, basedata, filename=args.out)

    elif args.vcfname is not None:
        filetype="VCF"
        if len(args.vcfname) > 1 and args.out_prefix is not None:
            raise Exception(("Multiple VCFs require --out-prefix flag"))
        for vcfname in args.vcfname:
            basedata = BaseData(args, "VCF",vcfname)
            intervals = getIntervalList(args,basedata)
            #sys.stderr.write(str(intervals)+'\n')
            interval_list.append(intervals)
            if args.out_prefix is not None:
                if args.returntype == 'returnlist':
                    for iv in intervals:
                        outputSubregion(args, iv, basedata)
                else:
                    outputSubregion(args, intervals, basedata)
            elif out_list:
                return interval_list
            elif args.out is not None:
                if args.returntype == 'returnlist':
                    raise Exception(('If --retl is specified, output must not '
                                    'be --out'))
                outputSubregion(args, intervals, basedata, filename=args.out)
    elif args.fasta is not None:
        filetype="FASTA"
        if len(args.fasta) > 1 and args.out_prefix is not None:
            raise Exception(("Multiple FASTAs require --out-prefix flag"))
        for fastaname in args.fasta:
            basedata = BaseData(args, "FASTA", fastaname)
            intervals = getIntervalList(args, basedata)
            interval_list.append(intervals)
    elif args.sit is not None:
        filetype="SIT"
        if len(args.sit) > 1 and args.out_prefix is not None:
            raise Exception(("Multiple SITES files require --out-prefix flag"))
        if args.out_reg is not None:
            raise Exception(("SITES files cannot be output to region file"))
        for sitname in args.sit:
            intervals = getIntervalList(args, "SIT", sitname)
            interval_list.append(intervals)




    return interval_list
        #


if __name__ == '__main__':
    initLogger()

    sample_fourgametetest_intervals(sys.argv[1:])




import sys
import argparse

def addRegionArgs(parser):
    parser.add_argument('--bed', dest="region_name",
                        help="Name of gene region file")
    parser.add_argument('--zero-ho', dest="zeroho", action="store_true",
                        help="Region list is 1 indexed, not 0")
    parser.add_argument('--zero-closed', dest="zeroclosed", action="store_true",
                        help="If set, use closed coordinates instead of half-open")
    parser.add_argument('--regcol', dest='colstr', help= (
                        "Comma-separated list of columns for gene region "
                        " data, format is start/end if no chromosome "
                        " data, start/end/chrom if so"))



import sys
import pysam
import argparse
import logging
import os
#sys.path.insert(0,os.path.abspath(os.path.join(os.pardir, 'andrew')))
from pgpipe.logging_module import initLogger
from pgpipe.genome_region import Region, RegionList
import pgpipe.vcf_reader_func as vf
from pgpipe.model import Model, read_model_file


def createParser():
    parser= argparse.ArgumentParser(description=("Given a range or a file of "
                                "ranges and a VCF file, will generate one or "
                                "more VCF files with variants only from the "
                                "region(s) specified. For regions, start and "
                                "end coordinates required (default is zero-"
                                "based, half-open intervals, chromosome "
                                "optional unless VCF has more than one"))
    parser.add_argument("vcfname", help="Input VCF name")
    region_group = parser.add_mutually_exclusive_group()
    region_group.add_argument("--r", dest="gene_str", help=("Comma "
                              "separated string, formatted \"start,end,"
                              "chrom\" or \"start,end\""))
    region_group.add_argument("--bed", dest="genename", help=("Tab-delimited"
                              " file with two (start-end) or three "
                              "(chromosome) columns"))
    parser.add_argument('--zero-ho', dest="zeroho", action="store_true")
    parser.add_argument('--zero-closed', dest="zeroclosed", action="store_true")
    parser.add_argument("--out", dest="output_name", help= (
                        "Optional name for output other than default"))
    parser.add_argument("--out-prefix",dest="out_prefix",help=("Specifies"
                        " output sent to multiple files starting with "
                        "given prefix"))
    parser.add_argument("--bed-column-index", dest="gene_col", help= (
                        "Comma-separated list of columns for gene region "
                        " data, format is start/end if no chromosome "
                        " data, start/end/chrom if so"))
    parser.add_argument("--compress-vcf", dest="compress_flag",
                        action="store_true", help=("If input VCF is not "
                        "compressed, will compress and use fetch search"))
    parser.add_argument("--parsecpg", dest="refname")
    parser.add_argument("--compress-out", dest="compress", action="store_true")
    parser.add_argument("--remove-indels",dest="remove_indels", 
                        action="store_true", 
                        help=("Removes indels from output VCF files"))
    parser.add_argument("--remove-multi", dest="remove_multiallele",
                        action="store_true")
    parser.add_argument("--remove-missing", dest="remove_missing", default=-1, 
                        help=("Will filter out site if more than the given number of "
                        "individuals (not genotypes) are missing data. 0 removes sites"
                        " with any missing data, -1 (default) removes nothing"))
    parser.add_argument("--informative-count", dest="informative_count",
                        type=int, default=0)
    parser.add_argument("--tbi", dest="tabix_index", help="Path to bgzipped "
                        "file's index if name doesn't match VCF file")
    parser.add_argument("--remove-missing-inds",dest="remove_missing_inds",
                        action="store_true", help=("Will remove individuals "
                        "with missing data from a loci's VCF file"))
    parser.add_argument("--multi-start-num",type=int,default=0,help=("If "
                        "multiple output files, start numbering with this number"))
    subsamp_group = parser.add_mutually_exclusive_group()
    subsamp_group.add_argument('--subsamp-list', dest="subsamp_fn",
                               help="List of sample names to be used")
    subsamp_group.add_argument('--subsamp-num', dest="subsamp_num",
                               help=("Number of individuals to be randomly "
                                     "subsampled from VCF file"))
    subsamp_group.add_argument('--model-file',dest="modelname",help="Model file for selecting individuals for writing")
    parser.add_argument('--model',dest="poptag",help="If model file is used, will use model with this name")
    parser.add_argument("--forceempty",dest="forceempty",action="store_true",help=("Will create empty VCF if a region is empty rather than throw an error"))
    return parser

def logArgs(args):
    logging.info('Arguments for vcf_region_write:')
    for k in vars(args):
        logging.info('Argument %s: %s' % (k, vars(args)[k]))

def getOutputName(args):
    vcfname = args.vcfname
    #nocompress = args.nocompress
    compress_output = args.compress
    if args.output_name is not None:
        if compress_output and args.output_name[-3:] != '.gz':
            return args.output_name+'.gz'
        return args.output_name
        #might need check for .gz if shouldnt be compressed
    end_ext = 'vcf.gz'

    if not compress_output:
        end_ext = 'vcf'
    for ext in ['vcf.gz','vcf']:
        offset = -1*len(ext)
        if ext == vcfname[offset:]:
            return args.vcfname[:offset]+'subregions.'+end_ext
    raise Exception(("Either --output needs a value or the vcf input needs "
                     "an extension of vcf or vcf.gz"))

def getOutputPrefix(args):
    if args.out_prefix is not None:
        return args.out_prefix
    if args.output_name is not None:
        return args.output_name
    for ext in ['vcf.gz','vcf']:
        offset = -1*len(ext)
        if ext == args.vcfname[offset:]:
            return args.vcfname[:offset]
    return args.vcfname

def getMultiFileName(pref, rc, compress):
    #oc = (not uncompressed and not nocompress_flag)
    ext = '.vcf'
    if compress:
        ext += '.gz'
    return pref+str(rc)+ext

def writeRegion(args, vcf_reader, region, rc, filter_sites, remove_cpg, 
                fasta_ref,header,vcf_out=None):
    rec_list = vcf_reader.getRecordList(region)
    out_p = getOutputPrefix(args)
    if len(rec_list) == 0:
        if args.forceempty:
            logging.warning(("Region %s has no variants "
                        "in VCF file") % (region.toStr()))
        else:
            raise Exception("Region %s has no variants" % (region.toStr()))
    outname = getMultiFileName(out_p,rc,args.compress)
    if vcf_out is None:

        vcf_t = pysam.VariantFile(outname,'w',header=header) #check
    else:
        vcf_t = vcf_out
    if filter_sites:
        pass_list = vf.getPassSites(rec_list, remove_cpg=remove_cpg,
                        remove_indels=args.remove_indels,
                        remove_multiallele=args.remove_multiallele,
                        remove_missing=args.remove_missing,
                        inform_level=args.informative_count,
                        fasta_ref=fasta_ref)
    for i in range(len(rec_list)):
        if (not filter_sites) or pass_list[i]:
            vcf_t.write(rec_list[i])
    if vcf_out is None:
        vcf_t.close()

def writeFile(args, vcf_reader, filter_sites, remove_cpg, fasta_ref, header):
    outname = getOutputName(args)
    outfile = pysam.VariantFile(outname,'w',header=header)
    for rec in vcf_reader.reader:
        if (not filter_sites) or vf.checkRecordPass(rec,
                                 remove_cpg=remove_cpg,
                                 remove_indels=args.remove_indels,
                                 remove_multiallele=args.remove_multiallele,
                                 remove_missing=args.remove_missing,
                                 inform_level=args.informative_count,
                                 fasta_ref=fasta_ref):
            outfile.write(rec)


def vcf_region_write(sys_args):
    """Returns a VCF file with variants from regions in a list

    Given a VCF file and gene region information (from either a file with one
    or more regions or a string with a single one), will output a subset of
    variants that occur in the given regions.

    Parameters
    ----------
    vcfname : str
        Filename for VCF input file. If it does not end with extension
        'vcf(.gz)', a value for --ext must be provided.
    --r : str, optional
        Comma-separates string with either two or three elements, the first
        two corresponding to the start and end coordinates of the target
        region. If there is a third element, it is taken as chromosome data.
        Must use either this or --rl flag. (Not both)
    --rl : str, optional
        Filename for gene region file. Requires columns for start and end
        coordinates, with option for chromosome. Additional data may be
        included, the columns with relevant data can be specified with the
        --gene-col option. Must use either this or --r flag. (Not both)
    --noindels : bool, optional
        If set, indels will not be included in the output VCF
    --output : str, optional
        If set, the default output name of (vcfinput).subregion.vcf(.gz) will
        be replaced with the given string. If writing multiple files, this
        value will be the file prefix, with outputs being of form
        (output).region(k).vcf[.gz]
    --nocompress
        If set, output VCF will be uncompressed. This will prevent adding
        the .gz extension to the output filename, which is how pysam
        determines what compression to use while writing
    --gene-col : str, optional
        Comma-separated string with two or three elements (chromosome is
        optional). If length is 2, elements are the indices for columns
        in the input gene region file corresponding to the start/end
        coordinates of a region. If length 3, the third element
        specifies the index of the chromosome column. Default is "1,2,0",
        to match column order in a BED file.



    Other Parameters
    ----------------
    --gr1 : bool, optional (False)
        If set, indicates that the genome coordinate data is in base 1
    --compress-vcf : bool, optional
        If set, will use bgzip and tabix to compress and index given VCF
        file
    --multi-out: bool, optional
        If set, each region will be written to a separate VCF file, tagged
        with (prefix).region[n].vcf(.gz)
    --subsamp-fn : str, optional
        Name of single-column file with names of individuals to subsample
        from input VCF file.
    --subsamp-num : int, optional
        Number of individuals to be randomly subsampled from VCF file

    output
    ------
    VCF file(s)
        Will be named either '--output' value or (vcfinput).subregion.vcf(.gz)
        Contains variants from regions specified in region list from input
        VCF.


    """

    parser = createParser()
    if len(sys_args) == 0:
        parser.print_help()
        sys.exit(1)
    args = parser.parse_args(sys_args)
    logArgs(args)
    popmodel = None
    if args.modelname is not None:
        popmodels = read_model_file(args.modelname)
        if len(popmodels) != 1:
            popmodel = popmodels[args.poptag]
        else:
            pp = list(popmodels.keys())
            popmodel = popmodels[pp[0]]
    vcf_reader = vf.VcfReader(args.vcfname,
                              compress_flag=args.compress_flag,
                              subsamp_num=args.subsamp_num,
                              subsamp_fn=args.subsamp_fn,
                              popmodel=popmodel,
                              index=args.tabix_index)
    logging.info('VCF file read')
    header = vcf_reader.reader.header
    #first_el = vcf_reader.prev_last_rec
    first_el = vcf_reader.info_rec
    chrom = first_el.chrom
    full_fileread = False
    vcf_out = None
    if args.out_prefix is None:
        output_name = getOutputName(args)
        vcf_out = pysam.VariantFile(output_name, 'w', header=header)

    if args.gene_str is not None:
        region_list = RegionList(genestr=args.gene_str,zeroho=args.zeroho,
                                 zeroclosed=args.zeroclosed)
    elif args.genename is not None:
        region_list = RegionList(filename=args.genename,zeroho=args.zeroho,
                                 zeroclosed=args.zeroclosed,
                                 colstr=args.gene_col,sortlist=False)
    else:
        full_fileread=True
    logging.info('Region read')

    fasta_ref = None
    remove_cpg = (args.refname is not None)
    filter_sites = ((args.refname is not None)
                   or args.remove_indels
                   or args.remove_multiallele
                   or (args.remove_missing != -1)
                   or (args.informative_count != 0))
    if args.refname is not None:
        fasta_ref = pysam.FastaFile(args.refname)
        remove_cpg = True

    logging.info('Total individuals: %d' % (len(first_el.samples)))

    if full_fileread:
        writeFile(args,vcf_reader,filter_sites,remove_cpg,fasta_ref,header)
    else:
        logging.info('Total regions: %d' % (len(region_list.regions)))
        for rc,region in enumerate(region_list.regions,start=args.multi_start_num):
            #if args.multi_out:
            writeRegion(args,vcf_reader,region,rc,filter_sites,remove_cpg,
                        fasta_ref,header,vcf_out)
                #continue
        #rec_list = vcf_reader.getRecordList(region)
        #if len(rec_list) == 0:
        #    if args.forceempty:
        #        logging.warning(("Region %s has no variants "
        #                    "in VCF file") % (region.toStr()))
        #    else:
        #        raise Exception("Region %s has no variants" % (region.toStr()))
        #if args.multi_out:
        #    try:
        #        vcf_out.close()
        #    except:
        #        pass
        #    outname = getMultiFileName(out_p, rc, args.compress)
        #    vcf_out = pysam.VariantFile(outname, 'w', header=header)
        #if filter_sites:
        #    pass_list = vf.getPassSites(rec_list, remove_cpg=remove_cpg,
        #                  remove_indels=args.remove_indels, 
        #                  remove_multiallele=args.remove_multiallele,
        #                  remove_missing=args.remove_missing,
        #                  inform_level=args.informative_count,
        #                  fasta_ref=fasta_ref)
        #for i in range(len(rec_list)):
            #make filter_sites an option
        #    if (not filter_sites) or pass_list[i]:
        #        vcf_out.write(rec_list[i])


if __name__ == '__main__':
    initLogger()
    vcf_region_write(sys.argv[1:])



import io
import pgpipe
from setuptools import setup
from Cython.Build import cythonize

with io.open("README.rst", "rt", encoding="utf8") as f:
    readme = f.read()

requirements = ['Cython',
                'pysam',
                'pandas',
                'pybedtools',
                'scikit-learn',
                'matplotlib']

ppp_scripts = ['pgpipe/vcf_filter.py',
               'pgpipe/vcf_calc.py',
               'pgpipe/vcf_split.py',
               'pgpipe/vcf_phase.py',
               'pgpipe/four_gamete.py',
               'pgpipe/stat_sampler.py',
               'pgpipe/bed_tasks.py',
               'pgpipe/informative_loci_filter.py',
               'pgpipe/convert.py',
               'pgpipe/vcf_to_ima.py',
               'pgpipe/admixture.py',
               'pgpipe/plink_ld.py',
               'pgpipe/ima3_wrapper.py',
               'pgpipe/eigenstrat_fstats.py',
               'pgpipe/vcf_liftover.py',
               'pgpipe/vcf_utilities.py']

setup(name=pgpipe.__title__,
      version=pgpipe.__version__,
      project_urls={"Documentation": "https://ppp.readthedocs.io",
                    "Code": "https://github.com/jaredgk/PPP",
                    "Issue tracker": "https://github.com/jaredgk/PPP/issues"},
      license=pgpipe.__license__,
      url=pgpipe.__url__,
      author=pgpipe.__author__,
      author_email=pgpipe.__email__,
      maintainer="Jared Knoblauch, " \
                 "Andrew Webb",
      maintainer_email="tug41380@temple.edu",
      description=pgpipe.__summary__,
      long_description=readme,
      include_package_data=True,
      packages=['pgpipe'],
      install_requires=requirements,
      scripts=ppp_scripts,
      python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*",
      ext_modules = cythonize('pgpipe/test_cython.pyx'))

#packages=setuptools.find_packages() to automate package finding?



import os
import sys
import json
import subprocess
import argparse
import logging
import itertools
import copy

import numpy as np

from collections import defaultdict, OrderedDict

# Import logging module
from pgpipe.logging_module import initLogger

class ModelFile(dict):
    def __init__(self, *arg, **kw):
        super(ModelFile, self).__init__(*arg, **kw)
        self.inds = []
        self.ind_file = ''
        self.exclude_file = ''

        if arg and self.confirm_model_instance(arg[1]):
            self.update_inds(arg[1])

    def __setitem__(self, *arg, **kw):
        super(ModelFile, self).__setitem__(*arg, **kw)

        if arg and self.confirm_model_instance(arg[1]):
            self.update_inds(model = arg[1])

    def __delitem__(self, key):
        super(ModelFile, self).__delitem__(key)
        self.update_inds()

    def confirm_model_instance (self, unknown):

        if isinstance(unknown, Model):

            return True

        else:

            return False

    def copy_model (self, src_model_name, new_model_name):

        src_model = super(ModelFile, self).__getitem__(src_model_name)

        src_model_copy = copy.deepcopy(src_model)

        src_model_copy.name = new_model_name

        super(ModelFile, self).__setitem__(new_model_name, src_model_copy)

    def rename_model (self, src_model_name, new_model_name):

        src_model = super(ModelFile, self).pop(src_model_name)

        src_model.name = new_model_name

        super(ModelFile, self).__setitem__(new_model_name, src_model)

    def update_inds (self, model = None):

        if self.confirm_model_instance(model):

            # Return error if inds is empty
            if not model.inds:
                raise IOError('No individuals found in %s.' % model.name)

            # Create a list of the unique individuals
            unique_inds = list(set(self.inds + model.inds))

        else:

            # Create an empty list for the unique individuals
            unique_inds = []

            # Loop the models in the file
            for model_in_file in super(ModelFile, self).values():

                # Create a list of the unique individuals
                unique_inds = list(set(unique_inds + model_in_file.inds))


        # Store the individuals
        self.inds = unique_inds

    def create_ind_file (self, file_ext = '', file_path = '', overwrite = False):
        # Assign the filename for the population file
        ind_filename = 'unique_individuals' + file_ext

        # If a path is assigned, create the file at the specified location
        if file_path:
            ind_filename = os.path.join(file_path, ind_filename)

        # Check if previous files should be overwriten
        if not overwrite:
            # Check if the file already exists
            if os.path.isfile(ind_filename):
                raise IOError('Individuals file exists.')

        # Create the population file
        ind_file = open(ind_filename, 'w')
        ind_file.write('%s\n' %'\n'.join(self.inds))
        ind_file.close()

        # Save the individuals filename
        self.ind_file = ind_filename

    def delete_ind_file (self):
        # Check if an individuals file was created
        if self.ind_file:

            # Delete the individuals file
            os.remove(self.ind_file)

            # Remove the filename
            self.ind_file = ''

    def create_exclude_ind_file (self, inds_to_include = [], file_ext = '', file_path = '', overwrite = False):
        # Assign the filename for the population file
        ind_filename = 'exclude_individuals' + file_ext

        # If a path is assigned, create the file at the specified location
        if file_path:
            ind_filename = os.path.join(file_path, ind_filename)

        # Check if previous files should be overwriten
        if not overwrite:
            # Check if the file already exists
            if os.path.isfile(ind_filename):
                raise IOError('Individuals file exists.')

        # Create exclude list by removing included individuals
        exclude_inds = list(set(self.inds) - set(inds_to_include))

        # Create the population file
        ind_file = open(ind_filename, 'w')
        ind_file.write('%s\n' %'\n'.join(exclude_inds))
        ind_file.close()

        # Save the individuals filename
        self.exclude_file = ind_filename

    def delete_exclude_ind_file (self):
        # Check if an individuals file was created
        if self.exclude_file:

            # Delete the individuals file
            os.remove(self.exclude_file)

            # Remove the filename
            self.exclude_file = ''

    def to_json (self):

        model_file_json = []

        for model_name, model_data in super(ModelFile, self).items():
            model_file_json.append(model_data.to_json())

        return model_file_json

class Model:
    def __init__ (self, name):
        self.name = name
        self.tree = ''
        self.pop_list = []
        self.ind_dict = defaultdict(list)
        self.nind = defaultdict(int)
        self.pop_files = []
        self.ind_file = ''

    def __str__ (self):
        return self.name

    @property
    def npop (self):
        return len(self.pop_list)

    @property
    def inds (self):
        return list(itertools.chain.from_iterable(self.ind_dict.values()))

    def assign_tree (self, tree):
        self.tree = str(tree)

    def assign_pop (self, pop, inds = []):
        self.pop_list.append(str(pop))
        if inds:
            self.ind_dict[pop] = [str(ind) for ind in inds]
        self.nind[pop] = len(self.ind_dict[pop])

    def sample_pop (self, pop, sample_size, with_replacements = False):

        # Confirm the pop is in the model
        if str(pop) not in self.pop_list:

            # Raise error if pop not found
            raise Exception('%s not found' % pop)

        # Confirm the sample size is an int
        try:

            sample_size = int(sample_size)

        except:

            # Raise error if sample_size not an int
            raise Exception('%s not int' % sample_size)

        # Check if the sample size is larger than the pop
        if int(sample_size) > self.nind[pop]:

            # Raise error if sample_size is larger
            raise Exception('%s is larger than %s' % (sample_size, pop))

        # Use numpy choice to randomly sample the pop
        sampled_inds = np.random.choice(self.ind_dict[pop], sample_size, replace = with_replacements)

        # Save the sampled inds as a list
        self.ind_dict[pop] = list(sampled_inds)

    def sample_pops (self, sample_size, with_replacements = False):

        # Confirm the sample size is an int
        try:

            sample_size = int(sample_size)

        except:

            # Raise error if sample_size not an int
            raise Exception('%s not int' % sample_size)

        # Loop each pop in the pop list
        for pop in self.pop_list:

            # Check if the sample size is larger than the pop
            if int(sample_size) > self.nind[pop]:

                # Raise error if sample_size is larger
                raise Exception('%s is larger than %s' % (sample_size, pop))

        # Loop each pop in the pop list, if no error raised
        for pop in self.pop_list:

            # Use numpy choice to randomly sample the pop
            sampled_inds = np.random.choice(self.ind_dict[pop], sample_size, replace = with_replacements)

            # Save the sampled inds as a list
            self.ind_dict[pop] = list(sampled_inds)

    def create_pop_files (self, file_ext = '', file_path = '', overwrite = False):
        for pop in self.pop_list:
            # Assign the filename for the population file
            pop_filename = pop + file_ext

            # If a path is assigned, create the file at the specified location
            if file_path:
                pop_filename = os.path.join(file_path, pop_filename)

            # Check if previous files should be overwriten
            if not overwrite:
                # Check if the file already exists
                if os.path.isfile(pop_filename):
                    raise IOError('Population file exists.')

            # Create the population file
            pop_file = open(pop_filename, 'w')
            pop_file.write('%s\n' %'\n'.join(self.ind_dict[pop]))
            pop_file.close()

            # Save the population filename
            self.pop_files.append(pop_filename)

        logging.info('Population files created for %s' % self)

    def delete_pop_files (self):
        # Check if pop files were created
        if len(self.pop_files) != 0:

            # Loop the created pop files
            for pop_file in self.pop_files:
                # Delete the pop file
                os.remove(pop_file)

            # Remove the filenames
            self.pop_files = []

            logging.info('Population files deleted for %s' % self)

    def create_ind_file (self, file_ext = '', file_path = '', overwrite = False):
        # Assign the filename for the population file
        ind_filename = 'individual.keep' + file_ext

        # If a path is assigned, create the file at the specified location
        if file_path:
            ind_filename = os.path.join(file_path, ind_filename)

        # Check if previous files should be overwriten
        if not overwrite:
            # Check if the file already exists
            if os.path.isfile(ind_filename):
                raise IOError('Individuals file exists.')

        # Create the population file
        ind_file = open(ind_filename, 'w')
        ind_file.write('%s\n' %'\n'.join(self.inds))
        ind_file.close()

        # Save the individuals filename
        self.ind_file = ind_filename

        logging.info('Individuals file created for %s' % self)

    def delete_ind_file (self):
        # Check if an individuals file was created
        if self.ind_file:

            # Delete the individuals file
            os.remove(self.ind_file)

            # Remove the filename
            self.ind_file = ''

            logging.info('Individuals file deleted for %s' % self)

    def to_json (self):

        model_json = OrderedDict()

        model_json['name'] = self.name

        pop_json = OrderedDict()

        for pop in self.pop_list:

            pop_json[pop] = OrderedDict()

            pop_json[pop]['indv'] = self.ind_dict[pop]

        model_json['pops'] = pop_json

        return model_json

    def return_pop (self, ind):

        # Loop the pops within the model
        for pop, inds in self.ind_dict.items():

            # Check if the ind belongs to the pop
            if ind in inds:

                return pop

        raise Exception('Individual (%s) not found within %s' % (ind, self.name))

def pops_not_in_model (model, pops):

    # Check if a pop in pops is not within model.pop_list
    pops_not_found = set(pops) - set(model.pop_list)

    # Check any pops were not found
    if pops_not_found:
        return pops_not_found

    return None

def read_model_file (filename):

    # Check that the file exists
    if not os.path.isfile(filename):
        raise IOError

    # Create ModelFile object
    models_to_return = ModelFile()

    # Check if using python 2 or 3
    if sys.version_info[0] == 2:
        # Open the model file in python 2
        model_file = open(filename, 'rU')
    else:
        # Open the model file in python 3
        model_file = open(filename, 'r', newline=None)

    # Parse the model file using the json reader
    models_dict = json.load(model_file)

    # List to store all unique individuals (i.e. individuals in all models)
    individual_list = []

    # Loop the parsed models
    for model_dict in models_dict:

        # Create the model
        model = Model(str(model_dict['name']))

        # Loop the populations in the model
        for pop, pop_dict in model_dict['pops'].items():

            # Convert all individuals names to str
            ind_list = [str(pop_ind) for pop_ind in pop_dict['inds']]

            # Assign the population ans it's individuals to the model
            model.assign_pop(str(pop), ind_list)

            # Assign the individuals to the unique individual list
            individual_list.extend(ind_list)

        # Remove duplicates from the unique individual list
        individual_list = list(set(individual_list))

        # Save the model
        models_to_return[str(model.name)] = model

    logging.info('Finished reading model file (%s)' % filename)

    # Return the models
    return models_to_return

def read_single_model(filename, popname = None):
    #Returns single model from file, assumes either only one
    #in file or popname is provided
    popmodels = read_model_file(filename)
    if popname is None:
        if len(popmodels) > 1:
            raise Exception("Model file %s has %d models, must specify which to use" % (filename,len(popmodels)))
        pp = list(popmodels.keys())
        return popmodels[pp[0]]
    else:
        if popname not in popmodels.keys():
            raise Exception("Model %s not found in model file %s" % (popname,filename))
        return popmodels[popname]

def write_model_file (model_file, filename, overwrite = False):

    # Check if the file is to be overwritten
    if not overwrite:

        # Check if the file exists
        if os.path.exists(filename):
            raise Exception('%s already exists' % filename)

    # Open the output file
    output_file = open(filename, 'w')

    # Write the json-formmated data to the output file
    output_file.write(json.dumps(model_file.to_json(), indent = 4))

    # Close the output file
    output_file.close()

    logging.info('Finished writing model file (%s)' % filename)



import sys
import argparse
import logging
import numpy as np

from pgpipe.model import Model, read_single_model

def readLocus(ima_f,pop_count):
    line = ima_f.readline()
    seqs = []
    seqs_per_pop = [int(i) for i in line.strip().split()[1:1+pop_count]]
    total_seqs = sum(seqs_per_pop)
    for i in range(total_seqs):
        line = ima_f.readline()
        seqs.append(line.strip().split()[1])
    return seqs,seqs_per_pop

def adjCoef(n):
    a = 0
    for i in range(1,max(2,int(n))):
        a += float(1)/float(i)
    return a

def wattersonPerPop(seqs,seqs_per_pop):
    seq_idx = 0
    watt_pop = []
    for i in range(len(seqs_per_pop)):
        varsites = 0
        ref_seq = seqs[seq_idx]
        for j in range(len(ref_seq)):
            for k in range(1,seqs_per_pop[i]):
                if ref_seq[j] != seqs[k+seq_idx][j]:
                    varsites += 1
                    break
        a = adjCoef(seqs_per_pop[i]/2)
        #print (varsites,a,seqs_per_pop[i])
        watt_pop.append(float(varsites)/float(a))
        seq_idx += seqs_per_pop[i]
    #print (watt_pop)
    #exit()
    return watt_pop

def statGeomean(locus_stats):
    means = []
    for i in range(len(locus_stats[0])):
        ll = np.array([float(1) if locus_stats[j][i] == 0 else locus_stats[j][i] for j in range(len(locus_stats))])
        logl = np.log(ll)
        means.append(np.exp(logl.sum()/len(logl)))
        #lmean = 0
        #for j in range(len(locus_stats)):
    return means


locus_stats = []

ima_f = open(str(sys.argv[1]),'r')

ima_f.readline()
line = ima_f.readline()
while line[0] == '#':
    line = ima_f.readline()
pop_count = int(line.strip())
line = ima_f.readline()
pop_names = line.strip().split()
if len(pop_names) != pop_count:
    raise Exception("Error parsing, length of population list is different from population count")
line = ima_f.readline()
if ':' in line:
    line = ima_f.readline()

locus_count = int(line.strip())
for i in range(locus_count):
    locus_data,seqs_per_pop = readLocus(ima_f,pop_count)
    locus_stats.append(wattersonPerPop(locus_data,seqs_per_pop))

gms = statGeomean(locus_stats)
print (gms)
ne = max(gms)
print (max(gms))
q = ne*5
t = ne*2
m = float(2)/float(ne)
print ("Q: ",q)
print ("T: ",t)
print ("M: ",m)


#!/usr/bin/env python
'''
    Automates the liftover function from picard for VCF-formatted files.

    ##################
    Command-line Usage
    ##################
    The VCF liftover function may be called using the following command:

    .. code-block:: bash
        
        python vcf_liftover.py

    *************
    Example usage
    *************
    Converting an hg19 VCF file **chr22.vcf.gz ** to hg38.

    .. code-block:: bash
        
        python vcf_liftover.py --vcf chr22.vcf.gz --chain hg19ToHg38.over.chain

    ############
    Dependencies 
    ############
    * `Picard <https://broadinstitute.github.io/picard/>`_
    
    ############################
    Input Command-line Arguments
    ############################
    **--vcf** *<input_filename>*
        Argument used to define the filename of the VCF file for liftover.
    **--chain** *<chain_filename>*
        Argument used to define the filename of the chain file required
        for liftover.
    **--ref** *<ref_filename>*
        Argument used to define the filename of the reference FASTA.
    **--ref-dict** *<dict_filename>*
        Argument used to define the filename of the reference dictionary, 
        if the naming scheme is non-standard.

    #############################
    Output Command-line Arguments
    #############################
    **--out** *<output_filename>*
        Argument used to define the complete output filename, overrides **--out-prefix**.
    **--out-prefix** *<output_prefix>*
        Argument used to define the output prefix (i.e. filename without file extension)
    **--out-format** *<vcf, vcf.gz, bcf>*
        Argument used to define the desired output format. Formats include: uncompressed 
        VCF (vcf); compressed VCF (vcf.gz) [default]; and BCF (bcf).
    **--reject** *<reject_filename>*
        Argument used to define the complete reject filename, overrides 
        **--reject-prefix**. The reject file containsrecords that could not be lifted 
        over.   
    **--reject-prefix** *<reject_prefix>*
        Argument used to define the reject prefix (i.e. filename without file extension)
    **--reject-format** *<vcf, vcf.gz, bcf>*
        Argument used to define the desired format of the reject output. Formats include: 
        uncompressed VCF (vcf); compressed VCF (vcf.gz); and BCF (bcf). If no format is
        given, the format defined in **--out-format** is used.
    **--overwrite**
        Argument used to define if previous output should be overwritten.

    ###############################
    Liftover Command-line Arguments
    ###############################
    **--liftover-min-matchs** *<match_float>*
        Argument used to define the minimum percentage required for liftover of a variant.
    **--picard-max-records** *<record_int>*
        Argument used to define the number of records to store in memory for picard.
    **--keep-index**
        Argument used to define if the output VCF index created by picard should be kept.
    **--picard-path** *<path>*
        Argument used to define the path to locate picard.jar.
'''

import os
import sys
import argparse
import itertools
import copy
import shutil
import logging

from pgpipe.logging_module import initLogger, logArgs
from pgpipe.vcf_reader_func import checkFormat
from pgpipe.picard import call_picard, get_ref_dictonary, check_ref_file_association, remove_picard_index
from pgpipe.fasta import check_format

def vcf_liftover_parser(passed_arguments):
    '''
    VCF Liftover Argument Parser

    Assign the parameters for VCF Liftover using argparse.

    Parameters
    ----------
    passed_arguments : list, optional
        Parameters passed by another function. sys.argv is used if
        not given. 

    Raises
    ------
    IOError
        If the input, or other specified files do not exist
    '''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError('%s not found' % value)
                setattr(args, self.dest, value)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    vcf_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

    # Input arguments.
    vcf_parser.add_argument('--vcf', help = "Defines the filename of the VCF", type = str, required = True, action = parser_confirm_file())
    vcf_parser.add_argument('--ref', help = 'Defines the filename of the reference fasta (target assembly)', type = str, required = True, action = parser_confirm_file())
    vcf_parser.add_argument('--ref-dict', help = 'Defines the filename of the reference fasta dictionary, if filename scheme differs', type = str, action = parser_confirm_file())
    vcf_parser.add_argument('--chain', help = 'Defines the filename of the chain file for liftover', type = str, required = True, action = parser_confirm_file())

    # Output file arguments
    vcf_parser.add_argument('--out', help = 'Defines the complete output filename, overrides --out-prefix', type = str)
    vcf_parser.add_argument('--out-prefix', help = 'Defines the output prefix (i.e. filename without file extension)', default = 'out')
    out_format_list = ['vcf', 'vcf.gz', 'bcf']
    out_format_default = 'vcf.gz'
    vcf_parser.add_argument('--out-format', metavar = metavar_list(out_format_list), help = 'Defines the desired output format', type = str, choices = out_format_list, default = out_format_default)

    # Reject file arguments
    vcf_parser.add_argument('--reject', help = 'Defines the complete reject filename, overrides --out-prefix', type = str)
    vcf_parser.add_argument('--reject-prefix', help = 'Defines the reject prefix (i.e. filename without file extension)', default = 'reject')
    reject_format_list = ['vcf', 'vcf.gz', 'bcf']
    vcf_parser.add_argument('--reject-format', metavar = metavar_list(reject_format_list), help = 'Defines the desired output format. If not defined, **--out-format** is used', type = str, choices = reject_format_list)
    
    # Optional arguments
    vcf_parser.add_argument('--liftover-min-match', help = 'Defines the minimum percentage required for liftover of a variant', type = float) 

    # General arguments
    vcf_parser.add_argument('--overwrite', help = "Overwrite previous output files", action = 'store_true')
    vcf_parser.add_argument('--keep-index', help = "Defines if the output VCF index created by picard should be kept", action = 'store_true')
    vcf_parser.add_argument('--picard-max-records', help = 'Defines the number of records to store in memory for picard', type = int)
    vcf_parser.add_argument('--picard-path', help = "Defines path to locate picard.jar", type = str)

    if passed_arguments:
        return vcf_parser.parse_args(passed_arguments)
    else:
        return vcf_parser.parse_args()

def assign_vcf_extension (filename):

    # Checks if the file is unzipped, bgzipped, or gzipped
    vcfname_format = checkFormat(filename)

    if vcfname_format == 'vcf':
        return '.vcf'

    elif vcfname_format == 'gzip' or vcfname_format == 'bgzip':
        return '.vcf.gz'

    elif vcfname_format == 'bcf':
        return '.bcf'

    else:
        raise Exception('%s - Unknown file format' % filename)

def assign_ref_extension (filename):

    # Checks if the file is unzipped, bgzipped, or gzipped
    ref_format = check_format(filename)

    if ref_format == 'fasta':
        return '.fa'

    elif ref_format == 'gzip':
        return '.fa.gz'

    else:
        raise Exception('%s - Unknown file format' % filename)

def run (passed_arguments = []):
    '''
    Liftover for VCF-formatted files

    This function uses the argparse-based function :py:func:`vcf_liftover_parser`
    to parse either sys.argv or passed_arguments to obtain the parameters below. 
    The parameters are then translated to their PICARD equivalent. Once all the 
    parameters are assigned, PICARD is called.

    Parameters
    ----------
    --vcf : str
        Filename of the VCF
    --ref : str
        Target assembly reference fasta
    --ref-dict
        Dictionary file of reference fasta
    --chain : str
        Chain file from origin assembly to target assembly
    --out : str
        Complete output filename, overrides --out-prefix
    --out-prefix : str
        Output prefix
    --out-format : str
        Desired output format
    --reject : str
        Complete reject filename. Overrides --reject-prefix
    --reject-prefix : str
        Reject prefix
    --reject-format : str
        Desired reject format, same as --out-format if not defined
    -- liftover-min-match : float
        Minimum percentage required for liftover of a variant
    --overwrite
        If previous output should be overwriten
    --keep-index
        Keep the output VCF index created by picard
    --picard-max-records : int
        Number of records to store in memory for picard
    --picard-path : str
        Defines path to locate picard.jar

    Raises
    ------
    IOError
        Output file already exists
    '''

    # Grab VCF arguments from command line
    vcf_args = vcf_liftover_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(vcf_args, func_name = 'vcf_liftover')

    # Check if a reference dictionary was not assigned
    if not vcf_args.ref_dict:

        # Get filename of ref dictionary, if found
        ref_dict = get_ref_dictonary(vcf_args.ref)

        # Check if the filename was found
        if not ref_dict:
            raise IOError('No dictionary associated with reference. Please create one using fasta_utilities')

        # Assign the ref dictionary filename
        vcf_args.ref_dict = ref_dict

    # Assign file extension for VCF input file
    ref_ext = assign_ref_extension(vcf_args.ref)

    # Assign the ref filename, in case of rename
    vcf_liftover_ref = copy.deepcopy(vcf_args.ref)

    # Assign the dict filename, in case of rename
    vcf_liftover_dict = copy.deepcopy(vcf_args.ref_dict)

    # Check if the reference and reference dictionary are not associated for picard
    if not check_ref_file_association(vcf_args.ref, vcf_args.ref_dict):

        # Edit the ref argument
        vcf_liftover_dict = vcf_liftover_ref + '.dict'

        # Rename the ref file
        os.rename(vcf_args.ref_dict, vcf_liftover_dict)

        # Edit the ref argument
        vcf_liftover_ref += ref_ext

        # Rename the ref file
        os.rename(vcf_args.ref, vcf_liftover_ref)

    # Assign file extension for VCF input file
    vcfname_ext = assign_vcf_extension(vcf_args.vcf)

    # Assign the vcf filename, in case of rename
    vcf_liftover_vcf = copy.deepcopy(vcf_args.vcf)

    # Confirm input has correct file extension
    if vcfname_ext not in vcf_args.vcf:

        # Edit the VCF argument
        vcf_liftover_vcf += vcfname_ext

        # Rename the file
        os.rename(vcf_args.vcf, vcf_liftover_vcf)

    # Create list to store arguments for liftover
    vcf_liftover_args = ['LiftoverVcf']

    # Assign the input VCF file
    vcf_liftover_args.append('I=' + vcf_liftover_vcf)

    # Assign the chain file
    vcf_liftover_args.append('C=' + vcf_args.chain)

    # Assign the reference FASTA file
    vcf_liftover_args.append('R=' + vcf_liftover_ref)

    # Check if a output name has been assigned
    if vcf_args.out:

        # Assign the output name
        vcf_liftover_out = vcf_args.out + os.extsep + vcf_args.out_format

    else:

        # Assign the output name
        vcf_liftover_out = vcf_args.out_prefix + os.extsep + vcf_args.out_format

    # Assign the output VCF file
    vcf_liftover_args.append('O=' + vcf_liftover_out)

    # Check if a specific reject format has not been specified
    if not vcf_args.reject_format:

        # Assign the output format to the reject format
        vcf_args.reject_format = vcf_args.out_format

    # Check if a output name has been assigned
    if vcf_args.reject:

        # Assign the output name
        vcf_liftover_reject = vcf_args.reject + os.extsep + vcf_args.reject_format

    else:

        # Assign the output name
        vcf_liftover_reject = vcf_args.reject_prefix + os.extsep + vcf_args.reject_format

    # Assign the output VCF file
    vcf_liftover_args.append('REJECT=' + vcf_liftover_reject)

    # Check if the user has assigned a record max
    if vcf_args.picard_max_records:

        # Assign the output VCF file
        #vcf_liftover_args.extend(['--MAX_RECORDS_IN_RAM', str(vcf_args.picard_max_records)])
        vcf_liftover_args.append('MAX_RECORDS_IN_RAM=' + str(vcf_args.picard_max_records))

    # Call picard with the arguments
    call_picard(vcf_args.picard_path, vcf_liftover_args, vcf_liftover_out)

    # Check if the index file created by picard should be removed
    if not vcf_args.keep_index:
        remove_picard_index(vcf_liftover_out, vcf_args.out_format)

    # Reverts the VCF input file to orginal filename
    if vcf_liftover_vcf != vcf_args.vcf:
        os.rename(vcf_liftover_vcf, vcf_args.vcf)

    # Reverts the ref files to orginal filenames
    if vcf_liftover_ref != vcf_args.ref:
        os.rename(vcf_liftover_ref, vcf_args.ref)
        os.rename(vcf_liftover_dict, vcf_args.ref_dict)

    # Check if a output name has been assigned and rename
    if vcf_args.out:
        os.rename(vcf_liftover_out, vcf_args.out)

    # Check if a output name has been assigned and rename
    if vcf_args.reject:
        os.rename(vcf_liftover_reject, vcf_args.reject)


if __name__ == "__main__":
    #initLogger()
    run()


import os
import sys
import subprocess
import shutil
import argparse
import glob
import logging

#sys.path.insert(0, os.path.abspath(os.path.join(os.pardir, 'pppipe')))

from pgpipe.vcf_reader_func import checkFormat
from pgpipe.logging_module import initLogger, logArgs
from pgpipe.vcftools import bgzip_decompress_vcfgz
from pgpipe.bcftools import convert_vcf, check_for_index, create_index

def delete_beagle_log (output_prefix):
    '''
        Delete beagle log file

        This function is used to delete beagle's log file if an error is
        encountered. A warning is produced if the log file cannot be found.

        Parameters
        ----------
        output_prefix : str
            Output file prefix
    '''

    # Check that log file exists, if not return warning
    if not os.path.isfile(output_prefix + '.log'):
        logging.warning('beagle log file %s.log does not exist' % output_prefix)
    else:
        os.remove(output_prefix + '.log')

def check_beagle_for_errors (beagle_stderr, output_prefix):
    '''
        Checks the beagle stdout for errors

        Parameters
        ----------
        beagle_stderr : str
            beagle stderr
        output_prefix : str
            Output file prefix

        Raises
        ------
        Exception
            If beagle stdout returns an error
    '''

    # Check if beagle completed without an error
    if not beagle_stderr.strip():
        pass

    # Print missing data message if that is likely
    elif 'ERROR: genotype is missing allele separator:' in str(beagle_stderr):
        # Delete the beagle log file
        delete_beagle_log(output_prefix)

        # Store reported error
        error_reported = 'ERROR: genotype is missing allele separator'
        # Store message for user about error
        user_message = 'Please confirm the input has no missing data.'
        # Report on the error
        raise Exception(error_reported + '\n' + user_message)

    # Print output for beagle if error is detected
    elif 'ERROR:' in str(beagle_stderr):
        # Delete the beagle log file
        delete_beagle_log(output_prefix)

        # Splits log into list of lines
        beagle_stderr_lines = beagle_stderr.splitlines()
        # Prints the error(s)
        raise Exception('\n'.join((output_line for output_line in beagle_stderr_lines if output_line.startswith('ERROR:'))))

    # Print output if not completed and no error found. Unlikely to be used, but included.
    else:
        # Delete the beagle log file
        delete_beagle_log(output_prefix)

        raise Exception(beagle_stderr)

def check_for_beagle_intermediate_files (output_prefix, output_format, overwrite = False):

    # Check if the output format is not vcf.gz
    if output_format != 'vcf.gz':

        # Check if the intermediate should be removed
        if not overwrite:

            # Check if an intermediate file exists and riase error
            if os.path.isfile(output_prefix + '.vcf.gz'):
                raise Exception('Beagle intermediate (%s) found. Use --overwrite to ignore')

def standard_beagle_call (beagle_path, beagle_call_args, output_prefix):
    '''
        Calls beagle using subprocess

        This function is used to call beagle under standard conditions. The
        functions then passes the stderr to check_beagle_for_errors to check
        for errors.

        Parameters
        ----------
        beagle_path : str
            Path to beagle.jar
        beagle_call_args : list
            Argument list for beagle
        output_prefix : str
            Output file prefix
    '''

    # Assign location of beagle jar file
    beagle_jar = os.path.join(beagle_path, 'beagle.jar')

    # Check that beagle.jar exists
    if not os.path.isfile(beagle_jar):
        raise IOError('beagle.jar not found. Path specified: %s' % beagle_path)

    logging.info('beagle phasing parameters assigned')

    # Phasing subprocess call
    phase_call = subprocess.Popen(['java', '-jar', beagle_jar] + beagle_call_args, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    phase_stdout, phase_stderr = phase_call.communicate()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        phase_stderr = phase_stderr.decode()

    # Check beagle call for errors
    check_beagle_for_errors(phase_stderr, output_prefix)

    logging.info('beagle phasing complete')

def call_beagle (beagle_path, beagle_call_args, output_prefix, output_format):
    '''
        Automates beagle calls

        This function passes the argument list to standard_beagle_call. Once the
        beagle call has finished, the function will automatically convert the
        bgzip compressed output of beagle to BCF and VCF, if either format is
        specified.

        Parameters
        ----------
        beagle_path : str
            Path to beagle.jar
        beagle_call_args : list
            Argument list for beagle
        output_prefix : str
            Output file prefix
        output_format : str
            Output file format
    '''

    # Standard call to beagle
    standard_beagle_call(beagle_path, beagle_call_args, output_prefix)

    # Check if the desired format is VCF
    if output_format == 'vcf':

        # Decompress the VCF file
        bgzip_decompress_vcfgz(output_prefix + '.vcf.gz')

    # Check if the desired format is BCF
    elif output_format == 'bcf':

        # Check if there is an index file
        if check_for_index(output_prefix + '.vcf.gz') == False:
            # Create an index if not found
            create_index(output_prefix + '.vcf.gz')
        
        # Convert vcf.gz to bcf
        convert_vcf(output_prefix + '.vcf.gz', output_prefix, output_format, overwrite = True)



import os
import sys
import subprocess
import shutil
import argparse
import glob
import logging
import pkg_resources

#sys.path.insert(0, os.path.abspath(os.path.join(os.pardir, 'pppipe')))

from pgpipe.vcf_reader_func import checkFormat
from pgpipe.logging_module import initLogger, logArgs
from pgpipe.plink import convert_haps_to_vcf

def check_shapeit_for_errors (shapeit_stdout, output_prefix):
    '''
        Checks the shapeit stdout for errors

        Parameters
        ----------
        shapeit_stdout : str
            shapeit stdout
        output_prefix : str
            Output filename prefix

        Raises
        ------
        Exception
            If shapeit stdout returns an error
    '''

    # Returns True if the job completed without error
    if 'Running time:' in str(shapeit_stdout):
        pass

    # Print output if not completed and no error found. Unlikely to be used, but included.
    else:
        # Remove intermediate files before reporting the error
        remove_shapeit_intermediate_files(output_prefix, error_intermediates = True)
        raise Exception(str(shapeit_stdout))

def remove_shapeit_intermediate_files (output_prefix, error_intermediates = False):
    '''
        Removes shapeit intermediate files

        This function is used to remove the various intermediate files created
        by shapeit. The exact intermediate files to be removed are defined by
        the error-state of shapeit. The function will also return warnings if
        the intermediate files were not found.

        Parameters
        ----------
        output_prefix : str
            Output filename prefix
        error_intermediates : bool, optional
            Defines if shapeit encountered an error

    '''
    if error_intermediates:

        # Check that the log file was created, give a warning otherwise
        if not os.path.isfile(output_prefix + '.phase.log'):
            logging.warning('shapeit intermediate file %s.phase.log does not exist' % output_prefix)
        else:
            # Remove shapeit log file
            os.remove(output_prefix + '.phase.log')

    else:

        # Check that the phase.ind.mm file was created, give a warning otherwise
        if not os.path.isfile(output_prefix + '.phase.ind.mm'):
            logging.warning('shapeit intermediate file %s.phase.ind.mm does not exist' % output_prefix)
        else:
            # Remove shapeit phase.ind.mm file
            os.remove(output_prefix + '.phase.ind.mm')

        # Check that the phase.snp.mm file was created, give a warning otherwise
        if not os.path.isfile(output_prefix + '.phase.snp.mm'):
            logging.warning('shapeit intermediate file %s.phase.snp.mm does not exist' % output_prefix)
        else:
            # Remove shapeit phase.snp.mm file
            os.remove(output_prefix + '.phase.snp.mm')

    # Check that the haps file was created, give a warning otherwise
    if not os.path.isfile(output_prefix + '.haps'):
        logging.warning('shapeit intermediate file %s.haps does not exist' % output_prefix)
    else:
        # Remove shapeit haps file
        os.remove(output_prefix + '.haps')

    # Check that the sample file was created, give a warning otherwise
    if not os.path.isfile(output_prefix + '.sample'):
        logging.warning('shapeit intermediate file %s.sample does not exist' % output_prefix)
    else:
        # Remove shapeit sample file
        os.remove(output_prefix + '.sample')

    logging.info('shapeit-related files removed')

def check_for_shapeit_intermediate_files (output_prefix, overwrite = False):

    # Check if intermediates files should not be overwritten
    if not overwrite:

        # List to hold intermediate files
        shapeit_intermediate_files = []
    
        # Check if a phase.ind.mm file exists
        if os.path.isfile(output_prefix + '.phase.ind.mm'):

            # Add the intermediate file to the list
            shapeit_intermediate_files.append(output_prefix + '.phase.ind.mm')

        # Check if a phase.snp.mm file exists
        if os.path.isfile(output_prefix + '.phase.snp.mm'):

            # Add the intermediate file to the list
            shapeit_intermediate_files.append(output_prefix + '.phase.snp.mm')

        # Check if a haps file exists
        if os.path.isfile(output_prefix + '.haps'):

            # Add the intermediate file to the list
            shapeit_intermediate_files.append(output_prefix + '.haps')

        # Check if a sample file exists
        if os.path.isfile(output_prefix + '.sample'):

            # Add the intermediate file to the list
            shapeit_intermediate_files.append(output_prefix + '.sample')

        # Check if intermediate_files were found, and report the error
        if shapeit_intermediate_files:
            raise Exception('shapeit intermediate files exist (%s)' % ', '.join(shapeit_intermediate_files))

def standard_shapeit_call (shapeit_call_args, output_prefix):
    '''
        Calls shapeit using subprocess

        This function is used to call shapeit and passes the resulting stdout
        to check_shapeit_for_errors to check for errors. The function also
        passes output_prefix to check_shapeit_for_errors to delete shapeit
        intermediate files if shapeit results in an error.

        Parameters
        ----------
        shapeit_call_args : list
            Argument list for shapeit
        output_prefix : str
            Output filename prefix

    '''

    logging.info('shapeit phasing parameters assigned')
    #shapeit_path = '../bin/shapeit'
    #shapeit_path = pkg_resources.resource_filename('shapeit','../bin/shapeit')
    # Phasing subprocess call
    #shapeit_path='shapeit'
    shapeit_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))),'bin','shapeit')

    #sys.stderr.write(str(shapeit_path)+'\n')
    phase_call = subprocess.Popen([shapeit_path] + shapeit_call_args, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    phase_stdout, phase_stderr = phase_call.communicate()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        phase_stdout = phase_stdout.decode()

    # Check shapeit call for errors
    check_shapeit_for_errors(phase_stdout, output_prefix)

    logging.info('shapeit phasing complete (HAPS format)')

def call_shapeit (shapeit_call_args, output_prefix, output_format):
    '''
        Calls shapeit and automates file conversions

        The function is used to call shapeit and also automates conversion to
        VCF, VCF.GZ, and BCF using plink2

        Parameters
        ----------
        shapeit_call_args : list
            Argument list for shapeit
        output_prefix : str
            Output filename prefix
        output_format : str
            Output file format

    '''

    # Standard call to beagle
    standard_shapeit_call(shapeit_call_args, output_prefix)

    # Convert haps-format to vcf
    convert_haps_to_vcf(output_prefix, output_format)

    logging.info('HAPS conversion to VCF complete')





def getAlleleCountArray(char * instr, int slen):
    cdef int i = 0;
    cdef int field_count = 0;
    cdef bint active_field = 0;
    cdef int ref_total = 0;
    cdef int alt_total = 0;
    cdef int miss_total = 0;
    cdef char v = 0
    cdef int d[256]
    for i in range(256): d[i] = 0
    for i in range(slen):
        v = instr[i]
        if v == '\t':
            field_count += 1
            if field_count >= 9:
                active_field = 1
        if v == ';':
            active_field = 0
        if v == '|' or v == '/':
            continue
        if active_field:
            d[v] += 1
            #if v == '0':
            #    ref_total += 1
            #elif v == '1':
            #    alt_total += 1
            #elif v == '.':
            #    miss_total += 1
    return d
    #print (d['0'],d['1'],'0')

def getAlleleCountArraySub(char * instr, int slen, list idxlist, int ilen):
    cdef int i = 0;
    cdef int field_count = 0;
    cdef int iidx = 0;
    cdef bint active_field = 0;
    cdef int d[256];
    cdef char v = 0;
    for i in range(256): d[i] = 0
    for i in range(slen):
        v = instr[i]
        if v == '\t':
            field_count += 1
            if field_count >= 9:
                if idxlist[iidx] == field_count-9:
                    active_field = 1;
                    iidx += 1
                else:
                    active_field = 0
        elif v == ';':
            active_field = 0
        elif v == '|' or v == '/':
            continue
        if active_field:
            d[v] += 1
    return d

         




try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser
import argparse
import sys

is_python3 = (sys.version_info[0] == 3)



def parseOption(val):
    """Given a value read from a config file, returns appropriately formatted
    python variable.

    -If `val` == 'None', returns None variable
    -If `val` == 'True', returns True
    -If `val` == 'False', returns False
    -If `val` is a float, returns float-casted val
    -If `val` is an int and not a float, returns an int-casted val
    -If none of these criteria are met, returns the input value

    Parameters
    ----------
    val : str
        String from dict created by config parser

    Output
    ------
    val : multi
        Parsed config value for use with argparse


    """
    if val == 'None':
        return None
    if val == 'True':
        return True
    if val == 'False':
        return False
    try:
        a = int(val)
        b = float(val)
        if a == b:
            return int(val)
        return float(val)
    except:
        pass
    return val

def getConfigFilename(arglist, flag='--conf'):
    """Reads argument list to determine if there is a config file provided.

    Used to work with argparse. If `flag` is present in the input array,
    the next argument is assumed to be the filename for the config file.

    Parameters
    ----------
    arglist : list
        List of all arguments to input function, from either sys.argv or a
        custom list
    flag : str (optional) ('--conf')
        Command-line flag to look for in arglist to signify next element
        is filename for config options

    Returns
    -------
    config_filename : str (None)
        Name of config file if 'flag' is present in arglist, otherwise
        defaults to None
    is_first : bool
        Will be true if `flag` is first element of arglist. Used for method
        to allow required args to be passed from config file.

    """
    config_filename = None
    is_first = False
    if flag in arglist:
        idx = arglist.index(flag)
        config_filename = arglist[idx+1]
        if idx == 0:
            is_first = True
        return config_filename,is_first
    return None,is_first


def defaultsDictForFunction(func_name, config_name):
    """Creates a dict with default values for a given function.

    Given an input config file and a function name, this function will
    read in all values from the given section of the config file, then
    remove the variables that are used in the config file to define
    parameters for multiple functions.

    Parameters
    ----------
    func_name : str
        Name of function section of config file. (May not necessarily
        be name of function in code)
    config_name : str
        Name of config file

    Returns
    -------
    d_func : dict
        Dictionary with all attributes for the given function. Values written
        as 'None' in config file will not be added, so normal defaults
        can be used.


    """
    if is_python3:
        config = ConfigParser.ConfigParser()
    else:
        config = ConfigParser.SafeConfigParser()
    config.read(config_name)
    d_vars = dict(config.items("DEFAULT"))
    d_hold = dict(config.items(func_name))
    d_func = {}
    for k in d_hold.keys():
        if k not in d_vars:
            val = parseOption(d_hold[k])
            if val is not None:
                d_func[k] = val
    return d_func

def makeRequiredList(argdict, reqlist):
    """Grabs required args from argument dict created by configparser.

    Generates a list of required parameters that will be passed in lieu of
    a regular arglist or sys.argv. Must be ordered in the same way they will
    be input to argparse.

    Parameters
    ----------
    argdict : dict
        Dictionary of arguments returned from defaultsDictForFunction.
    reqlist : list
        List of required arguments, in order that they should be passed to
        argparse.

    Returns
    -------
    req_arglist : list
        List of required argument values for given function, in order for
        passing to argparse

    Exceptions
    ----------
    Argument value in argdict for required arg is None



    """
    req_arglist = []
    for arg in reqlist:
        argval = argdict[arg]
        if argval is None:
            raise Exception("Value for required argument %s cannot be None"
                            % argval)
        req_arglist.append(argdict[arg])
    return req_arglist


def checkRequired(required_args, defaults):
    for req in required_args:
        if req in defaults and req[defaults] is not None:
            raise Exception(('Required argument %s has value present '
                            ' in config file (%s) and command line' %
                            (req, defaults[req])))


def getArgsWithConfig(parser, sys_args, required_args, func_name):
    config_name, req_included = getConfigFilename(sys_args)
    if config_name is not None:
        defaults = defaultsDictForFunction(func_name, config_name)
        if not req_included:
            checkRequired(required_args, defaults)
        parser.set_defaults(**defaults)
    if not req_included:
        return parser.parse_args(sys_args)
    else:
        req_args = makeRequiredList(defaults, required_args)
        return parser.parse_args(req_args)
